// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import 'package:matrix_sdk/src/event/event.dart';
import 'package:matrix_sdk/src/event/room/message_event.dart';
import 'package:matrix_sdk/src/event/room/state/aliases_change_event.dart';
import 'package:matrix_sdk/src/event/room/state/power_levels_change_event.dart';
import 'package:matrix_sdk/src/event/room/state/room_upgrade_event.dart';
import 'package:matrix_sdk/src/event/room/state/topic_change_event.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/room/room.dart';
import 'package:matrix_sdk/src/user.dart';
import 'package:meta/meta.dart';

import 'redaction_event.dart';
import 'state/room_avatar_change_event.dart';
import 'state/member_change_event.dart';
import 'state/room_name_change_event.dart';
import 'state/room_creation_event.dart';

abstract class RoomEvent extends Event {
  final RoomId roomId;
  final User sender;
  final DateTime time;

  @override
  final SentState sentState;

  RoomEvent(RoomEventArgs args)
      : roomId = args.roomId,
        sender = args.sender,
        time = args.time,
        sentState = args.sentState,
        super(args.id, args.unsigned);

  @override
  Map<String, dynamic> toJson() => super.toJson()
    ..addAll({
      'sender': sender.id.toString(),
      'origin_server_ts': time.millisecondsSinceEpoch
    });

  /// Create event from the given [json].
  static Future<RoomEvent> fromJson(
    RoomEventArgs args,
    Map<String, dynamic> json, {
    ProcessContext context,
  }) async {
    final type = Event.typeOf(json['type']);

    final finalArgs = await args.parse(
      json: json,
      context: context,
    );

    if (finalArgs == null) {
      return null;
    }

    RoomEvent event;
    switch (type) {
      case RoomNameChangeEvent:
        event = RoomNameChangeEvent.fromJson(finalArgs, json);
        break;
      case RoomAvatarChangeEvent:
        event = RoomAvatarChangeEvent.fromJson(finalArgs, json);
        break;
      case MemberChangeEvent:
        event = MemberChangeEvent.fromJson(finalArgs, json);
        break;
      case MessageEvent:
        event = MessageEvent.fromJson(finalArgs, json);
        break;
      case RedactionEvent:
        event = RedactionEvent.fromJson(finalArgs, json);
        break;
      case RoomCreationEvent:
        event = await RoomCreationEvent.fromJson(
          finalArgs,
          json,
          context: context,
        );
        break;
      case RoomUpgradeEvent:
        event = RoomUpgradeEvent.fromJson(finalArgs, json);
        break;
      case TopicChangeEvent:
        event = TopicChangeEvent.fromJson(finalArgs, json);
        break;
      case PowerLevelsChangeEvent:
        event = PowerLevelsChangeEvent.fromJson(finalArgs, json);
        break;
      case AliasesChangeEvent:
        event = AliasesChangeEvent.fromJson(finalArgs, json);
        break;
    }

    if (event != null &&
        json.containsKey('unsigned') &&
        json['unsigned']?.containsKey('redacted_because') == true) {
      final redactedBecause = json['unsigned']['redacted_because'];
      if (redactedBecause != null) {
        final redaction = RedactionEvent.fromJson(
          await finalArgs.parse(
            json: redactedBecause,
            context: context,
          ),
          redactedBecause,
        );

        final redacted = RedactedEvent.from(
          redaction: redaction,
          original: event,
        );

        event = redacted;
      }
    }

    if (event != null && context != null) {
      context.timeline[event.id] = event;
    }

    if (event is MemberChangeEvent) {
      context?.setUser(event.content.subject);
    }

    return event;
  }
}

class RoomEventArgs {
  final EventId id;
  final RoomId roomId;
  final User sender;
  final DateTime time;
  final UnsignedData unsigned;
  final SentState sentState;

  RoomEventArgs({
    this.id,
    this.roomId,
    this.sender,
    this.time,
    this.unsigned,
    this.sentState = SentState.sent,
  });

  static RoomEventArgs from(RoomEvent e) {
    return RoomEventArgs(
      id: e.id,
      roomId: e.roomId,
      sender: e.sender,
      time: e.time,
      unsigned: e.unsigned,
      sentState: e.sentState,
    );
  }

  RoomEventArgs copyWith({
    EventId id,
    RoomId roomId,
    User sender,
    DateTime time,
    UnsignedData unsigned,
    SentState sentState,
  }) {
    return RoomEventArgs(
      id: id ?? this.id,
      roomId: roomId ?? this.roomId,
      sender: sender ?? this.sender,
      time: time ?? this.time,
      unsigned: unsigned ?? this.unsigned,
      sentState: sentState ?? this.sentState,
    );
  }

  Future<RoomEventArgs> parse({
    @required Map<String, dynamic> json,
    ProcessContext context,
    EventId id,
    RoomId roomId,
    User sender,
    DateTime time,
    UnsignedData unsigned,
    SentState sentState,
  }) async {
    // These might be null
    id = id ?? EventId(json['event_id']);

    roomId = this.roomId ?? roomId;

    final originServerTs = json['origin_server_ts'];
    if (originServerTs != null) {
      time = DateTime.fromMillisecondsSinceEpoch(originServerTs);
    }

    var senderId;
    try {
      senderId = UserId(json['sender']);
    } on ArgumentError {
      return null;
    }

    final contextIsNull = context == null;

    sender = this.sender ??
        (!contextIsNull ? context.currentUsers[senderId] : null) ??
        await context?.members?.getStateOf(senderId, at: time) ??
        User(
          id: senderId,
          state: UserState(
            roomId: roomId,
            since: time,
          ),
        );

    var unsignedJson = json['unsigned'];
    if (unsigned == null && unsignedJson != null) {
      final transactionId = json['unsigned']['transaction_id'];

      if (transactionId != null) {
        unsigned = UnsignedData(transactionId: transactionId);
      } else {
        unsigned = null;
      }
    }

    return this.copyWith(
      id: id,
      roomId: roomId,
      sender: sender,
      time: time,
      unsigned: unsigned,
      sentState: sentState,
    );
  }
}
