// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import 'dart:convert';

import 'package:chopper/chopper.dart';
import 'package:http/http.dart' as http;
import 'package:matrix_sdk/src/api/client.dart';
import 'package:matrix_sdk/src/api/media.dart';
import 'package:matrix_sdk/src/exception.dart';

import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/store.dart';
import 'package:matrix_sdk/src/user.dart';
import 'package:matrix_sdk/src/util/chopper.dart';

import 'package:kiwi/kiwi.dart';
import 'package:url/url.dart';
import 'package:meta/meta.dart';

import 'device.dart';

/// Represents a Matrix homeserver. Also used as the main entry point
/// of the SDK.
class Homeserver {
  final http.Client _httpClient;
  final ChopperClient _chopper;
  // TODO: Make lazy
  ClientService get _client => _chopper.getService();
  MediaService get _media => _chopper.getService();

  final Url url;

  /// Dependency injection scope.
  /// @internal
  String scope;

  /// Returns a homeserver based on `.well-known` info.
  /// If there is no well-known info (404), just use [url].
  ///
  /// Queries `$url/.well-known/matrix/client` for information.
  ///
  /// Throws [WellKnownException] on failure.
  static Future<Homeserver> fromWellKnown(
    Url url, {
    http.Client httpClient,
  }) async {
    final wellKnownUrl = url.resolve('/.well-known/matrix/client');

    httpClient = httpClient ?? http.Client();
    final response = await httpClient.get(wellKnownUrl);

    if (response.statusCode == 404) {
      return Homeserver(url);
    }

    if (response.statusCode != 200) {
      throw WellKnownFailPromptException('Response status code is not 200');
    }

    if (response.body == null || response.body.isEmpty) {
      throw WellKnownFailPromptException('Response body is null or empty');
    }

    var body;
    try {
      body = json.decode(response.body);
    } on FormatException {
      throw WellKnownFailPromptException('Response body is not valid JSON');
    }

    final homeserverInfo = body['m.homeserver'];
    if (homeserverInfo == null) {
      throw WellKnownFailPromptException('m.homeserver key is missing');
    }

    String baseUrl = homeserverInfo['base_url'];
    if (baseUrl == null) {
      throw WellKnownFailPromptException('base_url key is missing or null');
    }

    if (baseUrl.endsWith('/')) {
      baseUrl = baseUrl.substring(0, baseUrl.length - 1);
    }

    try {
      final url = Url.parse(baseUrl);
      return Homeserver(url);
    } on FormatException {
      throw WellKnownFailErrorException('base_url is not valid');
    }

    // TODO: Validation
  }

  Homeserver(this.url, {http.Client httpClient})
      : _httpClient = httpClient,
        _chopper = createChopper(
          url: url,
          client: httpClient,
        ) {
    scope = url.toString() + hashCode.toString();

    Container().registerInstance(this, name: scope);
    Container().registerInstance(_client, name: scope);
    Container().registerInstance(_media, name: scope);
  }

  /// Logs in to this homeserver and returns a [LocalUser], which can be
  /// used for further operations. The [user] can be either a [UserId] or
  /// just a [Username]. The [store] is where all user data (rooms, events, etc)
  /// will be stored. Use the [device] to specify a device name or id for this
  /// login. Note that the `id` and `userId` are allowed to be null, they'll
  /// be filled automatically afterwards.
  Future<LocalUser> login(
    UserIdentifier user,
    String password, {
    @required Store store,
    Device device,
  }) async {
    final response = await _client.login(json.encode({
      'type': 'm.login.password',
      'identifier': user.toIdentifierJson(),
      'password': password,
      'device_id': device?.id?.toString(),
      'initial_device_display_name': device?.name,
    }));

    final body = json.decode(response.body);

    String accessToken = body['access_token'];
    final userId = UserId(body['user_id']);

    // If device was null,
    // create one with the id from the repsonse
    if (device == null) {
      device = Device(
        id: DeviceId(body['device_id']),
        userId: userId,
      );
    } else {
      device = device.copyWith(
        id: DeviceId(body['device_id']),
        userId: userId,
      );
    }

    // Get profile information of this user
    final profileResponse = await _client.profile(userId: userId.toString());

    var displayName, avatarUrl;
    if (profileResponse.isSuccessful) {
      final profile = json.decode(profileResponse.body);

      displayName = profile['displayname'];

      if (profile['avatar_url'] != null) {
        avatarUrl = Url.parse(profile['avatar_url']);
      }
    }

    final localUser = LocalUser(
      id: userId,
      name: displayName,
      avatarUrl: avatarUrl,
      store: store,
      homeserver: this,
      accessToken: accessToken,
      currentDevice: device,
      httpClient: _httpClient,
    );

    await store?.open(user: localUser);

    return localUser;
  }

  /// Checks whether the given [username] is available for
  /// registration on this [Homeserver].
  Future<bool> isUsernameAvailable(Username username) async {
    var response;

    try {
      response = await _client.registerAvailable(username.toString());
    } on UsernameInUseException {
      return false;
    }

    switch (response.statusCode) {
      case 200:
        return true;
      default:
        return false;
    }
  }

  /// Download content via this [Homeserver].
  Future<Stream<List<int>>> download(Url url) async {
    if (url == null) {
      throw ArgumentError.value(url, 'url', 'Must not be null');
    }

    if (url.scheme != 'mxc') {
      throw ArgumentError.value(url, 'url', 'Must be an mxc URL');
    }

    final server = url.authority;
    final mediaId = url.pathSegments[0];

    final response = await _media.download(server, mediaId);
    return response.body;
  }

  /// Download a thumbnail via this [Homeserver].
  Future<Stream<List<int>>> downloadThumbnail(
    Url url, {
    @required int width,
    @required int height,
    ResizeMethod resizeMethod = ResizeMethod.scale,
  }) async {
    if (url == null) {
      throw ArgumentError.value(url, 'url', 'Must not be null');
    }

    if (url.scheme != 'mxc') {
      throw ArgumentError.value(url, 'url', 'Must be an mxc URL');
    }

    final method = resizeMethod == ResizeMethod.scale ? 'scale' : 'crop';

    final server = url.authority;
    final mediaId = url.pathSegments[0];

    final response = await _media.thumbnail(
      server,
      mediaId,
      width,
      height,
      method,
    );
    return response.body;
  }
}

enum ResizeMethod {
  crop,
  scale,
}

/// Extracting information from `.well-known` failed. See
/// [message] for details.
abstract class WellKnownException implements Exception {
  final String message;

  WellKnownException(this.message);

  @override
  String toString() => '$runtimeType: $message';
}

/// Extracting information from `.well-known` failed. See
/// [message] for details.
///
/// Clients should `FAIL_PROMPT`, as per the spec.
class WellKnownFailPromptException extends WellKnownException {
  WellKnownFailPromptException(String message) : super(message);
}

/// Extracting information from `.well-known` failed. See
/// [message] for details.
///
/// Clients should `FAIL_ERROR`, as per the spec.
class WellKnownFailErrorException extends WellKnownException {
  WellKnownFailErrorException(String message) : super(message);
}
