// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.
import 'package:matrix_sdk/src/util/json.dart';
import 'package:validators/validators.dart';

import 'package:matrix_sdk/src/store.dart';

abstract class Id {
  final String value;

  Id(this.value);

  @override
  String toString() => value;

  String toJson() => toString();
}

/// A Matrix identifier.
abstract class MatrixId extends Id {
  MatrixId._(this.sigil, String value, {bool validate = true}) : super(value) {
    if (validate && !_isValidFullyQualified(value, sigil)) {
      throw ArgumentError.value(value, 'value',
          'Must be in format of: ${sigil}localpart${seperator}server.tld');
    }
  }

  final String sigil;
  static final String seperator = ':';

  @override
  String toString() => value;

  String get localPart {
    return value.split(seperator)[0].substring(1);
  }

  String get server {
    return value.split(seperator)[1];
  }

  bool operator ==(Object other) {
    if (other is Id) {
      return this.value == other.value;
    } else {
      return false;
    }
  }

  int get hashCode => value.hashCode;

  // Verification
  static final _localPartRegex = RegExp('^[A-z0-9\\-.=_\\/]+\$');

  static final _historicalLocalPartRegex = RegExp('^((?![:])[\x21-\x7F])*\$');

  /// Checks if it's a historically valid local part.
  ///
  /// Some old localparts have characters are now allowed because the spec
  /// in the past was more tolerant..
  static bool isValidHistoricalLocalPart(String input) =>
      _historicalLocalPartRegex.hasMatch(input);

  /// Check whether the given string is a valid
  /// local part of a user identifier.
  static bool isValidLocalPart(String input) => _localPartRegex.hasMatch(input);

  static bool _isValidFullyQualified(String input, String sigil) {
    if (input == null) {
      return false;
    }

    if (input.length > 255) {
      return false;
    }

    if (input.startsWith(sigil) && input.contains(seperator)) {
      var local, server;

      var seperatorIndex = input.indexOf(seperator);
      server = input.substring(seperatorIndex + 1);
      local = input.split(seperator)[0].substring(1);

      if (!isValidLocalPart(local) && !isValidHistoricalLocalPart(local)) {
        return false;
      }

      if (!isURL('https://$server')) {
        return false;
      }

      return true;
    } else {
      return false;
    }
  }

  static ValidId _isValid(String input, String sigil) {
    if (isValidLocalPart(input)) {
      return ValidId.local;
    }

    if (_isValidFullyQualified(input, sigil)) {
      return ValidId.full;
    }

    return ValidId.no;
  }
}

enum ValidId { no, local, full }

const _roomSigil = '!';

/// An internal room ID in the form `!something:server.tld`.
class RoomId extends MatrixId implements JsonWritable {
  RoomId(String value) : super._(_roomSigil, value);

  /// Checks whether the given string is a valid
  /// local part of a room id.
  ///
  /// Must match the regex `[a-z0-9\\-.=_\\/]+`
  static bool isValidLocalPart(String input) =>
      MatrixId.isValidLocalPart(input);

  /// Checks whether the given string is a valid
  /// fully qualified user id, must be in the form of:
  /// `!localpart:server.tld`.
  static bool isValidFullyQualified(String input) =>
      MatrixId._isValidFullyQualified(input, _roomSigil);

  /// Checks whether the given string is either a valid localpart (returns
  /// [ValidId.local]), a valid fully qualified room id (returns
  /// [ValidId.full]) or that it's not a valid id (returns [ValidId.no]).
  static ValidId isValid(String input) => MatrixId._isValid(input, _roomSigil);

  String toJson() => toString();
}

const _roomAliasSigil = '#';

class RoomAlias extends MatrixId implements JsonWritable {
  RoomAlias(String value) : super._(_roomAliasSigil, value);

  /// Checks whether the given string is a valid
  /// local part of a room id.
  ///
  /// Must match the regex `[a-z0-9\\-.=_\\/]+`
  static bool isValidLocalPart(String input) =>
      MatrixId.isValidLocalPart(input);

  /// Checks whether the given string is a valid
  /// fully qualified user id, must be in the form of:
  /// `!localpart:server.tld`.
  static bool isValidFullyQualified(String input) =>
      MatrixId._isValidFullyQualified(input, _roomAliasSigil);

  /// Checks whether the given string is either a valid localpart (returns
  /// [ValidId.local]), a valid fully qualified room id (returns
  /// [ValidId.full]) or that it's not a valid id (returns [ValidId.no]).
  static ValidId isValid(String input) =>
      MatrixId._isValid(input, _roomAliasSigil);
}

const _eventSigil = '\$';

/// An event ID in the form `$12323:server.tld`.
class EventId extends MatrixId implements JsonWritable {
  EventId(String value) : super._(_eventSigil, value, validate: false);

  String toJson() => toString();
}

const _userSigil = '@';

/// A Matrix ID in the form `@username:server.tld`.
class UserId extends MatrixId implements UserIdentifier, JsonWritable {
  UserId(String value) : super._(_userSigil, value);

  /// The [Username] of this ID, so in the case of `@joe:matrix.org`:
  /// `Username('joe')`.
  Username get username {
    return Username(localPart);
  }

  /// Checks whether the given string is a valid
  /// local part of a user identifier.
  ///
  /// Must match the regex `[a-z0-9\\-.=_\\/]+`
  static bool isValidLocalPart(String input) =>
      MatrixId.isValidLocalPart(input);

  /// Checks whether the given string is a valid
  /// fully qualified user id, must be in the form of:
  /// `@localpart:server.tld`.
  static bool isValidFullyQualified(String input) =>
      MatrixId._isValidFullyQualified(input, _userSigil);

  /// Checks whether the given string is either a valid localpart (returns
  /// [ValidId.local]), a valid fully qualified user id (returns
  /// [ValidId.full]) or that it's not a valid id (returns [ValidId.no]).
  static ValidId isValid(String input) => MatrixId._isValid(input, _userSigil);

  String toJson() => toString();

  /// Return this object represented as a user identifier, in the form
  /// of
  /// ```
  /// {
  ///   'type': 'm.id.user',
  ///   'user': '@joe:matrix.org'
  /// };
  /// ```
  @override
  Map<String, dynamic> toIdentifierJson() {
    return {
      'type': 'm.id.user',
      'user': '$value',
    };
  }
}

/// Localpart of a [UserId].
class Username implements UserIdentifier, JsonWritable {
  final String value;

  Username(this.value) {
    if (!MatrixId.isValidLocalPart(value)) {
      // throw
    }
  }

  bool operator ==(Object other) {
    if (other is Username) {
      return this.value == other.value;
    } else {
      return false;
    }
  }

  int get hashCode => value.hashCode;

  /// Checks whether the given string is a valid
  /// username.
  ///
  /// Must match the regex `[a-z0-9\\-.=_\\/]+`
  static bool isValid(String input) => MatrixId.isValidLocalPart(input);

  /// Return this object represented as a user identifier, in the form
  /// of
  /// ```
  /// {
  ///   'type': 'm.id.user',
  ///   'user': 'joe'
  /// };
  /// ```
  @override
  Map<String, dynamic> toIdentifierJson() {
    return {
      'type': 'm.id.user',
      'user': '$value',
    };
  }

  @override
  dynamic toJson() => value;

  @override
  String toString() => value;
}

/// Not to be confused with [UserId], the [UserIdentifier]
/// marks that the object can be used as a means to identify
/// a [User].
///
/// This could not only be a Matrix ID, but also a phone
/// number for example.
abstract class UserIdentifier {
  Map<String, dynamic> toIdentifierJson();
}

/// An object that can be identified with the [I] type of
/// identifier.
abstract class Identifiable<I extends Id> {
  I get id;

  bool operator ==(Object other) {
    if (other is Identifiable<I>) {
      return this.id == other.id;
    } else {
      return false;
    }
  }

  int get hashCode => id.hashCode;

  Future<String> getCustomData(Store store, String key) async {
    return await store.getObjectCustomData(id, key);
  }

  Future<void> setCustomData(Store store, String key, String value) async {
    await store.setObjectCustomData(id, key, value);
  }
}
