// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import 'dart:async';

import 'package:kiwi/kiwi.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/store.dart';
import 'package:meta/meta.dart';

/// [Identifiable] objects that can possibly be stored.
///
/// All methods that return a [FutureOr], will return a
/// [Future] only if items need to be retrieved from the [store] or
/// remotely. Subsequent calls (where the same items are retrieved)
/// will be from memory and therefore not a [Future].
abstract class Storables<I extends Id, T extends Identifiable<I>> {
  final Map<I, T> _memory = Map();

  final Store store;

  Storables(String scope) : store = Container().resolve(scope);

  @protected
  Iterable<T> getFromMemory() => _memory.values;

  @protected
  void setInMemory(Iterable<T> items) {
    _memory.addEntries((items.map((item) => MapEntry(item.id, item))));
  }

  @protected
  void setInMemoryIfAbsent(T item, T ifAbsent()) {
    _memory.putIfAbsent(item.id, ifAbsent);
  }

  @protected
  void removeInMemory(I id) {
    _memory.remove(id);
  }

  @protected
  Future<T> getSingleFromStore(I id);

  @protected
  Future<Iterable<T>> getFromStore({T from, int count});

  @protected
  Future<int> getSavedInStoreCount();

  @protected
  Future<void> setInStore(Iterable<T> value);

  @protected
  Future<void> removeInStore(I id) => null;

  @protected
  Future<Iterable<T>> getFromRemote({int count});

  @protected
  Iterable<T> sort(Iterable<T> items) => items;

  @protected
  bool isLastPossibleItem(T item) => false;

  @protected
  bool isMax(int count) => false;

  FutureOr<T> operator [](I id) {
    // If the room is in memory, return that one
    final fromMemory = _memory[id];
    if (fromMemory != null) {
      return fromMemory;
    }

    // Look in the store
    return getSingleFromStore(id);
  }

  FutureOr<Iterable<T>> get({int upTo, bool allowRemote = true}) {
    if (upTo != null) {
      return paginate(
        page: 0,
        pageSize: upTo,
        allowRemote: allowRemote,
      );
    } else {
      return getSavedInStoreCount().then((savedInStoreCount) {
        return paginate(
          page: 0,
          pageSize: savedInStoreCount,
          allowRemote: allowRemote,
        );
      });
    }
  }

  // TODO: Only sort when saving instead of getting
  FutureOr<Iterable<T>> paginate({
    @required int page,
    int pageSize = 30,
    bool allowRemote = true,
  }) {
    int start = page * pageSize;
    int end = (page + 1) * pageSize;

    final fullMemory = sort(getFromMemory());

    final memory = fullMemory.skip(start).take(pageSize);

    T last;
    try {
      last = fullMemory.last;
    } on StateError {
      last = null;
    }

    if (memory.length >= pageSize ||
        isLastPossibleItem(last) ||
        isMax(memory.length)) {
      return memory;
    }

    // Not enough items have been retrieved for the page,
    // get from store
    start += memory.length;
    return getFromStore(
      from: last,
      count: end - start,
    ).then((store) async {
      final combined = sort(memory.followedBy(store).toSet());
      setInMemory(store);

      T from;
      try {
        from = combined.last;
      } on StateError {
        from = null;
      }

      if (combined.length >= pageSize ||
          !allowRemote ||
          isLastPossibleItem(from) ||
          isMax(combined.length)) {
        return combined;
      }

      // Not enough items have been retrieved for the page,
      // get from remote
      start += store.length;
      var remote = sort(await getFromRemote(count: end - start));

      setInMemory(remote);
      await setInStore(remote);

      if (from != null) {
        var passedFrom = false;

        final filteredRemote = List<T>();
        for (final item in remote) {
          if (passedFrom) {
            filteredRemote.add(item);
          }

          if (item == from) {
            passedFrom = true;
          }
        }

        // If the filtered remote is empty, assume that the remote list
        // seamlessly joins with `combined`, otherwise
        // let remote be filtered
        if (filteredRemote.isNotEmpty) {
          remote = filteredRemote;
        }
      }

      return sort(combined.followedBy(remote).toSet());
    });
  }

  Future<void> addAll(Iterable<T> values) async {
    final nonNullValues = values.where((value) => value != null);

    setInMemory(nonNullValues);

    await setInStore(nonNullValues);
  }

  Future<void> remove(I id) async {
    removeInMemory(id);
    await removeInStore(id);
  }
}
