// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import 'package:matrix_sdk/src/event/room/room_event.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/room/room.dart';
import 'package:matrix_sdk/src/user.dart';

import 'event/event.dart';

/// Stores all data (rooms, users, events) somewhere.
abstract class Store {
  bool get isOpen;

  /// Will save the [user] info if provided.
  Future<bool> open({LocalUser user});
  Future<void> close();

  Future<void> delete();

  Future<String> getSyncToken();
  Future<void> setSyncToken(String value);

  Future<LocalUser> getLocalUser();
  Future<void> setLocalUser(LocalUser value);

  Future<Room> getRoom(RoomId id);
  Future<void> setRooms(Iterable<Room> rooms);

  /// Returned sorted alphabetically based on the room ID.
  ///
  /// See [Rooms].
  Future<Iterable<Room>> getRooms({Room from, int count});
  Future<int> getRoomCount();

  Future<RoomEvent> getRoomEvent(Room room, EventId id);
  Future<void> removeEvent(EventId id);

  /// Returned sorted based on [RoomEvent.time], newest first.
  ///
  /// See [Timeline].
  Future<Iterable<RoomEvent>> getRoomEvents(
    Room room, {
    RoomEvent from,
    int count,
  });
  Future<int> getRoomEventCount(RoomId roomId);
  Future<Iterable<RoomEvent>> getUnsentEvents(Room room);
  Future<void> addRoomEvents(Room room, Iterable<RoomEvent> events);

  Future<User> getMember(RoomId roomId, UserId id, {DateTime at});

  /// Returned sorted alphabetically based on User name or if null User id.
  ///
  /// See [Members].
  Future<Iterable<User>> getMembers(RoomId id, {User from, int count});
  Future<int> getMemberCount(RoomId roomId);
  Future<void> addMembers(RoomId roomId, Iterable<User> members);

  Future<void> setAppCustomData(String key, String value);
  Future<String> getAppCustomData(String key);

  Future<void> setObjectCustomData(Id objectId, String key, String value);
  Future<String> getObjectCustomData(Id objectId, String key);
}

class MemoryStore implements Store {
  bool get isOpen => true;

  LocalUser localUser;

  Future<LocalUser> getLocalUser() async => localUser;
  Future<void> setLocalUser(LocalUser value) async => localUser = value;

  String syncToken;
  Future<String> getSyncToken() async => syncToken;
  Future<void> setSyncToken(String value) async => syncToken = value;

  Future<bool> open({LocalUser user}) async {
    await setLocalUser(user);

    return true;
  }

  Future<void> close() => null;

  Future<void> delete() => null;

  final rooms = Map<RoomId, Room>();
  Future<Room> getRoom(RoomId id) async => rooms[id];
  Future<void> setRooms(Iterable<Room> rooms) async {
    for (final room in rooms) {
      this.rooms[room.id] = room;
    }
  }

  Future<Iterable<Room>> getRooms({Room from, int count}) async {
    final alphabetical =
        (Room a, Room b) => a.id.toString().compareTo(b.id.toString());
    final sortedRooms = this.rooms.values.toList()..sort(alphabetical);
    if (from == null) {
      return sortedRooms;
    }

    bool passedFrom = false;
    final rooms = List<Room>();
    for (final room in sortedRooms) {
      if (passedFrom) {
        rooms.add(room);
      }

      if (room == from) {
        passedFrom = true;
      }
    }

    return rooms..sort(alphabetical);
  }

  Future<int> getRoomCount() async => rooms.values.length;

  final timelines = Map<RoomId, Map<EventId, RoomEvent>>();
  Future<RoomEvent> getRoomEvent(Room room, EventId id) async {
    return timelines[room.id] != null ? timelines[room.id][id] : null;
  }

  Future<void> removeEvent(EventId id) async {
    for (final events in this.timelines.values) {
      events.remove(id);
    }
  }

  Future<Iterable<RoomEvent>> getRoomEvents(
    Room room, {
    RoomEvent from,
    int count,
  }) async {
    return timelines[room.id].values.where((event) {
      return event.time.millisecondsSinceEpoch <
          (from?.time?.millisecondsSinceEpoch ?? -1);
    });
  }

  Future<int> getRoomEventCount(RoomId roomId) async =>
      timelines[roomId].values.length;

  Future<Iterable<RoomEvent>> getUnsentEvents(Room room) async {
    return timelines[room.id]
        .values
        .where((event) => event.sentState == SentState.unsent);
  }

  Future<void> addRoomEvents(Room room, Iterable<RoomEvent> events) async {
    if (timelines[room.id] == null) {
      timelines[room.id] = Map();
    }

    for (final event in events) {
      timelines[room.id][event.id] = event;
    }
  }

  final membersPerRoom = Map<RoomId, Map<UserId, User>>();
  Future<User> getMember(RoomId roomId, UserId id, {DateTime at}) async {
    return membersPerRoom[roomId] != null ? membersPerRoom[roomId][id] : null;
  }

  Future<Iterable<User>> getMembers(RoomId id, {User from, int count}) async {
    final alphabetical = (User a, User b) {
      return (a.name ?? a.id.toString()).compareTo((b.name ?? b.id.toString()));
    };

    final members = membersPerRoom[id]?.values?.toList();

    if (members == null) {
      return Iterable.empty();
    }

    members.sort(alphabetical);

    if (from == null) {
      return members;
    }

    bool passedFrom = false;
    final membersToReturn = List<User>();
    for (final member in members) {
      if (passedFrom) {
        membersToReturn.add(member);
      }

      if (member == from) {
        passedFrom = true;
      }
    }

    return membersToReturn..sort(alphabetical);
  }

  Future<int> getMemberCount(RoomId roomId) async {
    return membersPerRoom[roomId].values.length;
  }

  Future<void> addMembers(RoomId roomId, Iterable<User> members) async {
    if (membersPerRoom[roomId] == null) {
      membersPerRoom[roomId] = Map();
    }

    for (final member in members) {
      membersPerRoom[roomId][member.id] = member;
    }
  }

  final customData = Map<String, String>();
  Future<void> setAppCustomData(String key, String value) async =>
      customData[key] = value;

  Future<String> getAppCustomData(String key) async => customData[key];

  final objectCustomData = Map<Id, Map<String, String>>();
  Future<void> setObjectCustomData(
    Id objectId,
    String key,
    String value,
  ) async {
    if (objectCustomData == null) {
      objectCustomData[objectId] = Map();
    }

    objectCustomData[objectId][key] = value;
  }

  Future<String> getObjectCustomData(Id objectId, String key) async =>
      objectCustomData[objectId][key];
}
