// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import 'dart:convert';

import 'package:chopper/chopper.dart';
import 'package:kiwi/kiwi.dart';
import 'package:matrix_sdk/src/api/access_token_interceptor.dart';
import 'package:matrix_sdk/src/api/client.dart';
import 'package:matrix_sdk/src/api/media.dart';
import 'package:matrix_sdk/src/event/room/room_event.dart';
import 'package:matrix_sdk/src/homeserver.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/room/room.dart';
import 'package:matrix_sdk/src/room/rooms.dart';
import 'package:matrix_sdk/src/store.dart';
import 'package:matrix_sdk/src/util/chopper.dart';
import 'package:meta/meta.dart';
import 'package:rxdart/rxdart.dart';
import 'package:url/url.dart';
import 'package:http/http.dart' as http;

import 'device.dart';
import 'exception.dart';
import 'notifications/pushers.dart';

class UserState implements Comparable<UserState> {
  UserId _userId;

  final RoomId roomId;
  final DateTime since;

  final String displayName;
  final Url avatarUrl;

  UserState({
    @required this.roomId,
    @required this.since,
    this.displayName,
    this.avatarUrl,
  });

  bool isStateOf(UserId userId) => _userId == userId;

  @override
  int compareTo(UserState other) => since.compareTo(other.since);

  @override
  bool operator ==(dynamic other) {
    if (other is UserState) {
      return this._userId == other._userId &&
          this.roomId == other.roomId &&
          this.since == other.since &&
          this.displayName == other.displayName &&
          this.avatarUrl == other.avatarUrl;
    }

    return false;
  }

  @override
  int get hashCode =>
      roomId.hashCode +
      since.hashCode +
      displayName.hashCode +
      avatarUrl.hashCode;
}

/// A User with at a specific [state].
class User extends Identifiable<UserId> implements Comparable<User> {
  @override
  final UserId id;

  final UserState state;

  User({@required this.id, this.state}) {
    state?._userId = id;
  }

  String get name => state?.displayName;
  Url get avatarUrl => state?.avatarUrl;

  factory User.fromState(UserState state) => User(
        id: state._userId,
        state: state,
      );

  /// Returns true when this user has same
  /// state as [user].
  bool hasSameStateAs(User user) => this.state == user.state;

  @override
  int compareTo(User other) => state.compareTo(other.state);
}

/// A user which is authenticated and can send messages, join rooms etc.
class LocalUser extends User {
  final Store _store;

  final Homeserver homeserver;
  final String accessToken;

  Rooms _rooms;
  Rooms get rooms => _rooms;

  Pushers _pushers;
  Pushers get pushers => _pushers;

  final Device currentDevice;

  bool _isLoggedOut = false;
  bool get isLoggedOut => _isLoggedOut;

  /// @internal
  final String scope;

  String _syncToken;

  bool _syncing = false;
  bool get isSyncing => _syncing;

  final ChopperClient _chopper;
  ClientService get _client => _chopper.getService();
  MediaService get _media => _chopper.getService();

  static Future<LocalUser> fromStore(Store store) async {
    final isOpen = await store.open();

    if (!isOpen) {
      return null;
    }

    return store.getLocalUser();
  }

  String _name;

  @override
  String get name => _name;

  @override
  final Url avatarUrl;

  LocalUser({
    @required UserId id,
    String name,
    this.avatarUrl,
    @required this.homeserver,
    @required this.accessToken,
    @required this.currentDevice,
    @required Store store,
    http.Client httpClient,
  })  : _chopper = createChopper(
            client: httpClient,
            url: homeserver.url,
            interceptors: [AccessTokenInterceptor(accessToken)]),
        _store = store,
        scope = homeserver.scope + id.toString(),
        _name = name,
        super(id: id) {
    Container().registerInstance(this, name: scope);
    Container().registerInstance(_client, name: scope);
    Container().registerInstance(store, name: scope);

    _rooms = Rooms(scope);
    Container().registerInstance(_rooms, name: scope);

    _pushers = Pushers(scope);
  }

  /// Set the display name of this user to the given value.
  Future<void> setName(String name) async {
    final response = await _client.profileSetDisplayName(
      userId: id.toString(),
      body: json.encode({
        'displayname': name,
      }),
    );

    MatrixException.throwFromResponse(response);

    if (response.isSuccessful) {
      _name = name;
      await _store.setLocalUser(this);
    }
  }

  /// Invalidates the access token of the user. Makes all
  /// [LocalUser] calls unusable.
  Future<void> logout() async {
    await _client.logout();
    _isLoggedOut = true;
  }

  /// Send all unsent messages still in the [Store].
  Future<void> sendAllUnsent() async {
    for (Room room in await rooms.get()) {
      if (room is JoinedRoom) {
        for (RoomEvent event in await _store.getUnsentEvents(room)) {
          await for (final _ in room.send(
            event.content,
            transactionId: event.unsigned?.transactionId,
          )) {}
        }
      }
    }
  }

  Future<Url> upload({
    @required Stream<List<int>> byteStream,
    @required int length,
    @required String contentType,
    String fileName,
  }) async {
    if (byteStream == null) {
      throw ArgumentError.value(
        byteStream,
        'byteStream',
        'byteStream must not be null',
      );
    }

    final response = await _media.upload(
      byteStream,
      length.toString(),
      contentType,
      fileName,
    );

    if (response.statusCode == 200) {
      try {
        final body = json.decode(response.body);
        return Url.parse(body['content_uri']);
      } on FormatException {
        return null;
      }
    }

    return null;
  }

  // Sync //
  final _syncSubject = PublishSubject<SyncState>();
  Observable<SyncState> get sync => _syncSubject.stream;

  /// Syncs data with the user's [homeserver]. Use the [sync] [Observable] to
  /// listen to sync changes.
  ///
  /// If you `await` this method, it will wait until the sync has been stopped
  /// by [stopSync]. In most cases you would want to call it like this:
  /// `unawaited(user.startSync())`.
  Future<void> startSync({
    Duration maxRetryAfter = const Duration(seconds: 30),
  }) async {
    if (isLoggedOut) {
      throw StateError('The user can not be logged out');
    }

    _syncing = true;

    // This var is used to implements exponential backoff
    // until it reaches maxRetryAfter
    var retryAfter = 500;

    while (_syncing) {
      final state = await syncOnce(timeout: Duration(seconds: 10));
      _syncSubject.add(state);

      if (!state.succeeded) {
        await Future.delayed(Duration(milliseconds: retryAfter));
        retryAfter = (retryAfter * 1.5).floor();
        if (retryAfter > maxRetryAfter.inMilliseconds) {
          retryAfter = maxRetryAfter.inMilliseconds;
        }
      } else {
        // reset exponential backoff
        retryAfter = 500;
      }
    }
  }

  Future<SyncState> syncOnce({
    bool updateSyncToken = true,
    Map<String, dynamic> filter,
    timeout = Duration.zero,
    bool fullState = false,
  }) async {
    if (isLoggedOut) {
      throw StateError('The user can not be logged out');
    }

    _syncToken = await _store?.getSyncToken() ?? _syncToken;

    if (filter == null) {
      filter = Map<String, dynamic>();
    }

    if (filter['room'] == null) {
      filter['room'] = Map<String, dynamic>();
    }

    if (filter['room']['state'] == null) {
      filter['room']['state'] = Map<String, dynamic>();
    }

    filter['room']['state']['lazy_load_members'] = true;

    try {
      Response response = await _client.sync(
          since: _syncToken,
          fullState: fullState,
          filter: json.encode(filter),
          timeout: timeout.inMilliseconds);
      final body = json.decode(response.body);

      final changedRooms = await _processRooms(body, body['account_data']);

      if (updateSyncToken) {
        // Store sync token
        _syncToken = body['next_batch'] as String;
        await _store?.setSyncToken(_syncToken);
      }

      return SyncState(dirtyRooms: changedRooms);
    } on Exception catch (e) {
      return SyncState(exception: e);
    }
  }

  void stopSync() {
    _syncing = false;
  }

  // Returns list of changed rooms
  Future<List<Room>> _processRooms(
    Map<String, dynamic> body,
    Map<String, dynamic> accountData,
  ) async {
    final jRooms = body['rooms'];

    final processed = List<Room>();

    const join = 'join';
    const leave = 'leave';

    process(Map<String, dynamic> jsonRooms, {@required String type}) async {
      if (jsonRooms != null) {
        for (final entry in jsonRooms.entries) {
          final roomId = RoomId(entry.key);
          final json = entry.value;

          Room room = await rooms[roomId];
          if (room == null) {
            switch (type) {
              case join:
                room = JoinedRoom(id: roomId, scope: scope);
                break;
              case leave:
                room = LeftRoom(id: roomId, scope: scope);
                break;
            }
          } else {
            if (type == leave) {
              room = LeftRoom(
                id: room.id,
                scope: scope,
                stateEvents: room.stateEvents,
                directUser: room.directUser,
                previousBatch: room.previousBatch,
              );
            }
          }

          await room.processSync(json);
          await room.processUserAccountData(accountData);
          await room.processEphemeral(json);

          processed.add(room);
        }
      }
    }

    await process(jRooms[join], type: join);
    await process(jRooms[leave], type: leave);

    // Save rooms
    await rooms.addAll(processed);

    return processed;
  }

  // End Sync //
}

class SyncState {
  final bool succeeded;

  /// If [succeeded] is false, this is the exception that was thrown.
  final Exception exception;

  /// Rooms that have changed since the last sync (new events,
  /// typing notifactions, etc).
  final List<Room> dirtyRooms;

  SyncState({this.exception, this.dirtyRooms = const []})
      : succeeded = exception == null;
}
